from freeplay.errors import FreeplayConfigurationError
from freeplay.record import DefaultRecordProcessor
from freeplay.support import CallSupport
from freeplay.thin.resources.prompts import Prompts
from freeplay.thin.resources.recordings import Recordings
from freeplay.thin.resources.sessions import Sessions
from freeplay.thin.resources.test_runs import TestRuns


class Freeplay:
    def __init__(
            self,
            freeplay_api_key: str,
            api_base: str
    ) -> None:
        if not freeplay_api_key or not freeplay_api_key.strip():
            raise FreeplayConfigurationError("Freeplay API key not set. It must be set to the Freeplay API.")

        self.call_support = CallSupport(freeplay_api_key, api_base,
                                        DefaultRecordProcessor(freeplay_api_key, api_base))
        self.freeplay_api_key = freeplay_api_key
        self.api_base = api_base

        # Resources ========
        self.sessions = Sessions()
        self.prompts = Prompts(self.call_support)
        self.test_runs = TestRuns(self.call_support)
        self.recordings = Recordings(self.call_support)


from dataclasses import dataclass
from typing import Optional

from .errors import FreeplayConfigurationError


@dataclass
class OpenAIConfig:
    api_key: str
    base_url: Optional[str] = None

    def validate(self) -> None:
        if not self.api_key or not self.api_key.strip():
            raise FreeplayConfigurationError("OpenAI API key not set. It must be set to make calls to the service.")


@dataclass
class AzureConfig(OpenAIConfig):
    engine: Optional[str] = None
    api_version: Optional[str] = None

    def validate(self) -> None:
        super().validate()

        if self.api_version is None:
            raise FreeplayConfigurationError(
                "OpenAI API version not set. It must be set to make calls to the service.")

        if self.engine is None:
            raise FreeplayConfigurationError("Azure engine is not set. It must be set to make calls to the service.")


@dataclass
class AnthropicConfig:
    api_key: str


@dataclass
class ProviderConfig:
    anthropic: Optional[AnthropicConfig] = None
    openai: Optional[OpenAIConfig] = None
    azure: Optional[AzureConfig] = None

    def validate(self) -> None:
        if all(config is None for config in [self.anthropic, self.openai, self.azure]):
            FreeplayConfigurationError("At least one provider key must be set in ProviderConfig.")

        if self.openai is not None:
            self.openai.validate()



import json
import logging
from dataclasses import dataclass
from typing import Any, Dict, Generator, List, Optional, Tuple, Union

from .completions import (
    PromptTemplates,
    CompletionResponse,
    CompletionChunk,
    ChatCompletionResponse,
    ChatMessage
)
from .errors import FreeplayConfigurationError
from .flavors import Flavor, ChatFlavor, require_chat_flavor, get_chat_flavor_from_config
from .llm_parameters import LLMParameters
from .model import InputVariables
from .provider_config import ProviderConfig
from .record import (
    RecordProcessor,
    DefaultRecordProcessor
)
from .support import CallSupport

logger = logging.getLogger(__name__)
default_tag = 'latest'


class Session:
    def __init__(
            self,
            call_support: CallSupport,
            session_id: str,
            prompts: PromptTemplates,
            flavor: Optional[Flavor],
            provider_config: ProviderConfig,
            tag: str = default_tag,
            test_run_id: Optional[str] = None
    ) -> None:
        self.tag = tag
        self.call_support = call_support
        self.session_flavor = flavor
        self.provider_config = provider_config
        self.session_id = session_id
        self.prompts = prompts
        self.test_run_id = test_run_id

    def get_completion(
            self,
            template_name: str,
            variables: InputVariables,
            flavor: Optional[Flavor] = None,
            **kwargs: Any
    ) -> CompletionResponse:
        completion_flavor = flavor or self.session_flavor
        return self.call_support.prepare_and_make_call(self.session_id,
                                                       self.prompts,
                                                       template_name,
                                                       variables,
                                                       completion_flavor,
                                                       self.provider_config,
                                                       self.tag,
                                                       self.test_run_id,
                                                       completion_parameters=LLMParameters(kwargs))

    def get_completion_stream(
            self,
            template_name: str,
            variables: InputVariables,
            flavor: Optional[Flavor] = None,
            **kwargs: Any
    ) -> Generator[CompletionChunk, None, None]:
        completion_flavor = flavor or self.session_flavor
        return self.call_support.prepare_and_make_call_stream(self.session_id,
                                                              self.prompts,
                                                              template_name,
                                                              variables,
                                                              completion_flavor,
                                                              self.provider_config,
                                                              self.tag,
                                                              self.test_run_id,
                                                              completion_parameters=LLMParameters(kwargs))


class ChatSession(Session):
    def __init__(
            self,
            call_support: CallSupport,
            session_id: str,
            prompts: PromptTemplates,
            flavor: Optional[ChatFlavor],
            provider_config: ProviderConfig,
            template_name: str,
            variables: InputVariables,
            tag: str = default_tag,
            test_run_id: Optional[str] = None,
            messages: Optional[List[ChatMessage]] = None,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None,
    ) -> None:
        super().__init__(call_support, session_id, prompts, flavor, provider_config, tag, test_run_id)
        # A Chat Session tracks the template_name and variables for a set of chat completions.
        # Assumes these will be the same for subsequent chat messages.
        self.message_history = messages or []
        self.variables = variables
        self.metadata = metadata
        self.target_template = self.call_support.find_template_by_name(self.prompts, template_name)
        self.flavor = get_chat_flavor_from_config(flavor, self.target_template.flavor_name)
        self.__initial_messages = json.loads(self.flavor.format(self.target_template, self.variables))

    def last_message(self) -> Optional[ChatMessage]:
        return self.message_history[len(self.message_history) - 1]

    def store_new_messages(self, new_messages: List[ChatMessage]) -> None:
        for message in new_messages:
            self.message_history.append({
                "role": message["role"],
                "content": message["content"]
            })

    def start_chat(self, **kwargs: Any) -> ChatCompletionResponse:
        response = self.call_support.prepare_and_make_chat_call(
            self.session_id,
            flavor=self.flavor,
            provider_config=self.provider_config,
            tag=self.tag,
            test_run_id=self.test_run_id,
            target_template=self.target_template,
            variables=self.variables,
            message_history=self.__initial_messages,
            new_messages=None,
            completion_parameters=LLMParameters(kwargs),
            metadata=self.metadata,
        )

        self.store_new_messages(response.message_history)
        return response

    def start_chat_stream(self, **kwargs: Any) -> Generator[CompletionChunk, None, None]:
        return self.continue_chat_stream(new_messages=None, **kwargs)

    def aggregate_message_from_response(
            self,
            response: Generator[CompletionChunk, None, None]
    ) -> Generator[CompletionChunk, Any, None]:
        message: ChatMessage = {
            "role": "assistant",
            "content": ""
        }

        for chunk in response:
            message["content"] += chunk.text
            yield chunk

        self.message_history.append(message)

    def continue_chat(
            self,
            new_messages: Optional[List[ChatMessage]] = None,
            **kwargs: Any
    ) -> ChatCompletionResponse:

        response = self.call_support.prepare_and_make_chat_call(
            self.session_id,
            flavor=self.flavor,
            provider_config=self.provider_config,
            tag=self.tag,
            test_run_id=self.test_run_id,
            target_template=self.target_template,
            variables=self.variables,
            message_history=self.message_history,
            new_messages=new_messages,
            completion_parameters=LLMParameters(kwargs),
            metadata=self.metadata,
        )

        if new_messages is not None:
            self.store_new_messages(new_messages)
        if response.content:
            self.message_history.append(response.message_history[-1])

        return response

    def continue_chat_stream(
            self,
            new_messages: Optional[List[ChatMessage]] = None,
            **kwargs: Any
    ) -> Generator[CompletionChunk, None, None]:
        new_messages = new_messages or []
        if len(self.message_history) == 0:
            self.message_history = self.__initial_messages

        response = self.call_support.prepare_and_make_chat_call_stream(
            self.session_id,
            flavor=self.flavor,
            provider_config=self.provider_config,
            tag=self.tag,
            target_template=self.target_template,
            variables=self.variables,
            message_history=self.message_history,
            test_run_id=self.test_run_id,
            completion_parameters=LLMParameters(kwargs),
            metadata=self.metadata,
        )

        self.store_new_messages(new_messages)
        yield from self.aggregate_message_from_response(response)


@dataclass()
class FreeplayTestRun:
    def __init__(
            self,
            call_support: CallSupport,
            flavor: Optional[Flavor],
            provider_config: ProviderConfig,
            test_run_id: str,
            inputs: List[InputVariables]
    ):
        self.call_support = call_support
        self.flavor = flavor
        self.provider_config = provider_config
        self.test_run_id = test_run_id
        self.inputs = inputs

    def get_inputs(self) -> List[InputVariables]:
        return self.inputs

    def create_session(self, project_id: str, tag: str = default_tag) -> Session:
        session_id = self.call_support.create_session_id()
        prompts = self.call_support.get_prompts(project_id, tag)
        return Session(self.call_support, session_id, prompts, self.flavor, self.provider_config,
                       tag, self.test_run_id)


# This SDK prototype does not support full functionality of either OpenAI's API or Freeplay's
# The simplifications are:
#  - Always assumes there is a single choice returned, does not support multiple
#  - Does not support an "escape hatch" to allow use of features we don't explicitly expose


class Freeplay:
    def __init__(
            self,
            freeplay_api_key: str,
            api_base: str,
            provider_config: ProviderConfig,
            flavor: Optional[Flavor] = None,
            record_processor: Optional[RecordProcessor] = None,
            **kwargs: Any
    ) -> None:
        if not freeplay_api_key or not freeplay_api_key.strip():
            raise FreeplayConfigurationError("Freeplay API key not set. It must be set to use the Freeplay API.")
        provider_config.validate()

        self.__record_processor = record_processor or DefaultRecordProcessor(freeplay_api_key, api_base)
        self.call_support = CallSupport(freeplay_api_key, api_base, self.__record_processor, **kwargs)
        self.provider_config = provider_config
        self.client_flavor = flavor
        self.freeplay_api_key = freeplay_api_key
        self.api_base = api_base

    def create_session(self, project_id: str, tag: str = default_tag) -> Session:
        session_id = self.call_support.create_session_id()
        prompts = self.call_support.get_prompts(project_id, tag)
        return Session(
            self.call_support,
            session_id,
            prompts,
            self.client_flavor,
            self.provider_config,
            tag)

    def restore_session(
            self,
            project_id: str,
            session_id: str,
            template_name: str,
            variables: InputVariables,
            tag: str = default_tag,
            flavor: Optional[Flavor] = None,
            **kwargs: Any
    ) -> CompletionResponse:
        prompts = self.call_support.get_prompts(project_id, tag)
        completion_flavor = flavor or self.client_flavor
        return self.call_support.prepare_and_make_call(
            session_id=session_id,
            prompts=prompts,
            template_name=template_name,
            variables=variables,
            flavor=completion_flavor,
            provider_config=self.provider_config,
            tag=tag,
            completion_parameters=LLMParameters(kwargs),
        )

    def get_completion(
            self,
            project_id: str,
            template_name: str,
            variables: InputVariables,
            tag: str = default_tag,
            flavor: Optional[Flavor] = None,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None,
            **kwargs: Any
    ) -> CompletionResponse:
        self.call_support.check_all_values_string_or_number(metadata)
        session_id = self.call_support.create_session_id()
        prompts = self.call_support.get_prompts(project_id, tag)
        completion_flavor = flavor or self.client_flavor

        return self.call_support.prepare_and_make_call(session_id,
                                                       prompts,
                                                       template_name,
                                                       variables,
                                                       completion_flavor,
                                                       self.provider_config,
                                                       tag,
                                                       completion_parameters=LLMParameters(kwargs),
                                                       metadata=metadata)

    def get_completion_stream(
            self,
            project_id: str,
            template_name: str,
            variables: InputVariables,
            tag: str = default_tag,
            flavor: Optional[Flavor] = None,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None,
            **kwargs: Any
    ) -> Generator[CompletionChunk, None, None]:
        self.call_support.check_all_values_string_or_number(metadata)
        session_id = self.call_support.create_session_id()
        prompts = self.call_support.get_prompts(project_id, tag)
        completion_flavor = flavor or self.client_flavor

        return self.call_support.prepare_and_make_call_stream(session_id,
                                                              prompts,
                                                              template_name,
                                                              variables,
                                                              completion_flavor,
                                                              self.provider_config,
                                                              tag,
                                                              completion_parameters=LLMParameters(kwargs),
                                                              metadata=metadata)

    def start_chat(
            self,
            project_id: str,
            template_name: str,
            variables: InputVariables,
            tag: str = default_tag,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None,
            **kwargs: Any
    ) -> Tuple[ChatSession, ChatCompletionResponse]:
        session = self.__create_chat_session(project_id, tag, template_name, variables, metadata)
        completion_response = session.start_chat(**kwargs)
        return session, completion_response

    def restore_chat_session(
            self,
            project_id: str,
            template_name: str,
            session_id: str,
            variables: InputVariables,
            tag: str = default_tag,
            messages: Optional[List[ChatMessage]] = None,
            flavor: Optional[ChatFlavor] = None) -> ChatSession:
        prompts = self.call_support.get_prompts(project_id, tag)
        chat_flavor = flavor or require_chat_flavor(self.client_flavor) if self.client_flavor else None
        return ChatSession(
            call_support=self.call_support,
            session_id=session_id,
            prompts=prompts,
            flavor=chat_flavor,
            provider_config=self.provider_config,
            template_name=template_name,
            variables=variables,
            tag=tag,
            messages=messages or []
        )

    def start_chat_stream(
            self,
            project_id: str,
            template_name: str,
            variables: InputVariables,
            tag: str = default_tag,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None,
            **kwargs: Any
    ) -> Tuple[ChatSession, Generator[CompletionChunk, None, None]]:
        """Returns a chat session, the base prompt template messages, and a streamed response from the LLM."""
        session = self.__create_chat_session(project_id, tag, template_name, variables, metadata)
        completion_response = session.start_chat_stream(**kwargs)
        return session, completion_response

    def create_test_run(self, project_id: str, testlist: str) -> FreeplayTestRun:
        test_run_response = self.call_support.create_test_run(project_id=project_id, testlist=testlist)

        return FreeplayTestRun(
            self.call_support,
            self.client_flavor,
            self.provider_config,
            test_run_response.test_run_id,
            [test_case.variables for test_case in test_run_response.test_cases]
        )

    def __create_chat_session(
            self,
            project_id: str,
            tag: str,
            template_name: str,
            variables: InputVariables,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None) -> ChatSession:
        chat_flavor = require_chat_flavor(self.client_flavor) if self.client_flavor else None

        session_id = self.call_support.create_session_id()
        prompts = self.call_support.get_prompts(project_id, tag)
        return ChatSession(
            self.call_support,
            session_id,
            prompts,
            chat_flavor,
            self.provider_config,
            template_name,
            variables,
            tag,
            metadata=metadata)



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, TypedDict

from openai.types.chat.chat_completion_chunk import ChoiceDeltaFunctionCall

from .llm_parameters import LLMParameters


class ChatMessage(TypedDict):
    role: str
    content: str


class OpenAIFunctionCall(TypedDict):
    name: str
    arguments: str


@dataclass
class CompletionResponse:
    content: str
    is_complete: bool
    openai_function_call: Optional[OpenAIFunctionCall] = None


@dataclass
class ChatCompletionResponse:
    content: str
    is_complete: bool
    message_history: List[ChatMessage]


@dataclass
class PromptTemplateWithMetadata:
    prompt_template_id: str
    prompt_template_version_id: str

    name: str
    content: str
    flavor_name: Optional[str]
    params: Optional[Dict[str, Any]]

    def get_params(self) -> LLMParameters:
        return LLMParameters.empty() if self.params is None else LLMParameters(self.params)


@dataclass
class PromptTemplates:
    templates: List[PromptTemplateWithMetadata]


@dataclass
class CompletionChunk:
    text: str
    is_complete: bool
    openai_function_call: Optional[ChoiceDeltaFunctionCall] = None



import json
from abc import abstractmethod, ABC
from copy import copy
from typing import cast, Any, Dict, Generator, List, Optional, Union

import anthropic
import openai
from openai import AuthenticationError, BadRequestError, Stream
from openai.types.chat import ChatCompletion, ChatCompletionChunk, ChatCompletionMessageParam

from .completions import CompletionChunk, PromptTemplateWithMetadata, CompletionResponse, ChatCompletionResponse, \
    ChatMessage, OpenAIFunctionCall
from .errors import FreeplayConfigurationError, LLMClientError, LLMServerError, FreeplayError
from .llm_parameters import LLMParameters
from .model import InputVariables
from .provider_config import AnthropicConfig, AzureConfig, OpenAIConfig, ProviderConfig
from .utils import bind_template_variables


class Flavor(ABC):
    @classmethod
    def get_by_name(cls, flavor_name: str) -> 'Flavor':
        if flavor_name == OpenAIChat.record_format_type:
            return OpenAIChat()
        elif flavor_name == AzureOpenAIChat.record_format_type:
            return AzureOpenAIChat()
        elif flavor_name == AnthropicClaudeChat.record_format_type:
            return AnthropicClaudeChat()
        else:
            raise FreeplayConfigurationError(
                'Configured flavor not found in SDK. Please update your SDK version or configure '
                'a different model in the Freeplay UI.')

    @property
    @abstractmethod
    def provider(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def record_format_type(self) -> str:
        raise NotImplementedError()

    @property
    def _model_params_with_defaults(self) -> LLMParameters:
        return LLMParameters.empty()

    @abstractmethod
    def format(self, prompt_template: PromptTemplateWithMetadata, variables: InputVariables) -> str:
        pass

    @abstractmethod
    def to_llm_syntax(self, messages: List[ChatMessage]) -> str | List[ChatMessage]:
        raise NotImplementedError()

    @abstractmethod
    def call_service(
            self,
            formatted_prompt: str,
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> CompletionResponse:
        pass

    @abstractmethod
    def call_service_stream(
            self,
            formatted_prompt: str,
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> Generator[CompletionChunk, None, None]:
        pass

    def get_model_params(self, llm_parameters: LLMParameters) -> LLMParameters:
        return self._model_params_with_defaults.merge_and_override(llm_parameters)


class ChatFlavor(Flavor, ABC):
    @abstractmethod
    def continue_chat(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> ChatCompletionResponse:
        pass

    @abstractmethod
    def continue_chat_stream(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> Generator[CompletionChunk, None, None]:
        pass


class OpenAIChatFlavor(ChatFlavor, ABC):

    @abstractmethod
    def _call_openai(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters,
            stream: bool
    ) -> Union[ChatCompletion, openai.Stream[ChatCompletionChunk]]:
        pass

    def format(self, prompt_template: PromptTemplateWithMetadata, variables: InputVariables) -> str:
        # Extract messages JSON to enable formatting of individual content fields of each message. If we do not
        # extract the JSON, current variable interpolation will fail on JSON curly braces.
        messages_as_json: List[Dict[str, str]] = json.loads(prompt_template.content)
        formatted_messages = [
            {
                "content": bind_template_variables(message['content'], variables), "role": message['role']
            } for message in messages_as_json]
        return json.dumps(formatted_messages)

    def to_llm_syntax(self, messages: List[ChatMessage]) -> List[ChatMessage]:
        return messages

    def call_service(
            self,
            formatted_prompt: str,
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> CompletionResponse:
        messages = json.loads(formatted_prompt)
        completion = cast(ChatCompletion, self._call_openai(messages, provider_config, llm_parameters, stream=False))

        return CompletionResponse(
            content=completion.choices[0].message.content or '',
            is_complete=completion.choices[0].finish_reason == 'stop',
            openai_function_call=self.__maybe_function_call(completion),
        )

    # noinspection PyMethodMayBeStatic
    def __maybe_function_call(self, completion: ChatCompletion) -> Optional[OpenAIFunctionCall]:
        maybe_function_call = completion.choices[0].message.function_call
        if maybe_function_call:
            return OpenAIFunctionCall(
                name=maybe_function_call.name,
                arguments=maybe_function_call.arguments
            )
        return None

    def call_service_stream(
            self,
            formatted_prompt: str,
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> Generator[CompletionChunk, None, None]:
        messages = json.loads(formatted_prompt)
        completion_stream = cast(Stream[ChatCompletionChunk],
                                 self._call_openai(messages, provider_config, llm_parameters, stream=True))
        for chunk in completion_stream:
            yield CompletionChunk(
                text=chunk.choices[0].delta.content or '',
                is_complete=chunk.choices[0].finish_reason == 'stop',
                openai_function_call=chunk.choices[0].delta.function_call
            )

    def continue_chat(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> ChatCompletionResponse:
        completion = cast(ChatCompletion, self._call_openai(messages, provider_config, llm_parameters, stream=False))

        message_history = copy(messages)
        message = completion.choices[0].message
        message_history.append(ChatMessage(
            role=message.role or '',
            content=message.content or ''
        ))
        return ChatCompletionResponse(
            content=message.content or '',
            message_history=message_history,
            is_complete=completion.choices[0].finish_reason == "stop"
        )

    def continue_chat_stream(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> Generator[CompletionChunk, None, None]:
        completion_stream = cast(Stream[ChatCompletionChunk],
                                 self._call_openai(messages, provider_config, llm_parameters, stream=True))
        for chunk in completion_stream:
            yield CompletionChunk(
                text=chunk.choices[0].delta.content or '',
                is_complete=chunk.choices[0].finish_reason == "stop"
            )


class OpenAIChat(OpenAIChatFlavor):
    record_format_type = "openai_chat"
    _model_params_with_defaults = LLMParameters({
        "model": "gpt-3.5-turbo"
    })

    def __init__(self) -> None:
        self.client: Optional[openai.OpenAI] = None

    @property
    def provider(self) -> str:
        return "openai"

    def get_openai_client(self, openai_config: Optional[OpenAIConfig]) -> openai.OpenAI:
        if self.client:
            return self.client

        if not openai_config:
            raise FreeplayConfigurationError(
                "Missing OpenAI key. Use a ProviderConfig to specify keys prior to getting completion.")

        self.client = openai.OpenAI(api_key=openai_config.api_key, base_url=openai_config.base_url)
        return self.client

    def _call_openai(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters,
            stream: bool
    ) -> Union[ChatCompletion, openai.Stream[ChatCompletionChunk]]:
        client = self.get_openai_client(provider_config.openai)
        try:
            return client.chat.completions.create(
                messages=cast(List[ChatCompletionMessageParam], messages),
                **self.get_model_params(llm_parameters),
                stream=stream,
            )
        except (BadRequestError, AuthenticationError) as e:
            raise LLMClientError("Unable to call OpenAI") from e
        except Exception as e:
            raise LLMServerError("Unable to call OpenAI") from e


class AzureOpenAIChat(OpenAIChatFlavor):
    record_format_type = "azure_openai_chat"

    def __init__(self) -> None:
        self.client: Optional[openai.AzureOpenAI] = None

    @property
    def provider(self) -> str:
        return "azure"

    def get_azure_client(
            self,
            azure_config: Optional[AzureConfig],
            api_version: Optional[str] = None,
            endpoint: Optional[str] = None,
            deployment: Optional[str] = None,
    ) -> openai.AzureOpenAI:
        if self.client:
            return self.client

        if not azure_config:
            raise FreeplayConfigurationError(
                "Missing Azure key. Use a ProviderConfig to specify keys prior to getting completion.")

        self.client = openai.AzureOpenAI(
            api_key=azure_config.api_key,
            api_version=api_version,
            azure_endpoint=endpoint or '',
            azure_deployment=deployment,
        )
        return self.client

    def _call_openai(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters,
            stream: bool
    ) -> Any:
        api_version = llm_parameters.get('api_version')
        deployment_id = llm_parameters.get('deployment_id')
        resource_name = llm_parameters.get('resource_name')
        endpoint = f'https://{resource_name}.openai.azure.com'
        llm_parameters.pop('resource_name')

        client = self.get_azure_client(
            azure_config=provider_config.azure,
            api_version=api_version,
            endpoint=endpoint,
            deployment=deployment_id,
        )

        try:
            return client.chat.completions.create(
                messages=cast(List[ChatCompletionMessageParam], messages),
                **self.get_model_params(llm_parameters),
                stream=stream,
            )
        except (BadRequestError, AuthenticationError) as e:
            raise LLMClientError("Unable to call Azure") from e
        except Exception as e:
            raise LLMServerError("Unable to call Azure") from e


class AnthropicClaudeChat(ChatFlavor):
    record_format_type = "anthropic_chat"
    _model_params_with_defaults = LLMParameters({
        "model": "claude-2",
        "max_tokens_to_sample": 100
    })

    def __init__(self) -> None:
        self.client: Optional[anthropic.Anthropic] = None

    @property
    def provider(self) -> str:
        return "anthropic"

    def get_anthropic_client(self, anthropic_config: Optional[AnthropicConfig]) -> anthropic.Client:
        if self.client:
            return self.client

        if not anthropic_config:
            raise FreeplayConfigurationError(
                "Missing Anthropic key. Use a ProviderConfig to specify keys prior to getting completion.")

        self.client = anthropic.Client(api_key=anthropic_config.api_key)
        return self.client

    # This just formats the prompt for uploading to the record endpoint.
    # TODO: Move this to a base class.
    def format(self, prompt_template: PromptTemplateWithMetadata, variables: InputVariables) -> str:
        # Extract messages JSON to enable formatting of individual content fields of each message. If we do not
        # extract the JSON, current variable interpolation will fail on JSON curly braces.
        messages_as_json: List[Dict[str, str]] = json.loads(prompt_template.content)
        formatted_messages = [
            {
                "content": bind_template_variables(message['content'], variables),
                "role": self.__to_anthropic_role(message['role'])
            } for message in messages_as_json]
        return json.dumps(formatted_messages)

    def to_llm_syntax(self, messages: List[ChatMessage]) -> str:
        formatted_messages = [
            ChatMessage(
                content=message['content'],
                role=self.__to_anthropic_role(message['role'])
            ) for message in messages
        ]
        return self.__to_anthropic_chat_format(formatted_messages)

    @staticmethod
    def __to_anthropic_role(role: str) -> str:
        if role == 'Human':
            return 'Human'
        elif role == 'assistant' or role == 'Assistant':
            return 'Assistant'
        else:
            # Anthropic does not support system role for now.
            return 'Human'

    @staticmethod
    def __to_anthropic_chat_format(messages: List[ChatMessage]) -> str:
        formatted_messages = []
        for message in messages:
            formatted_messages.append(f"{message['role']}: {message['content']}")
        formatted_messages.append('Assistant:')

        return "\n\n" + "\n\n".join(formatted_messages)

    def continue_chat(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> ChatCompletionResponse:
        formatted_prompt = self.__to_anthropic_chat_format(messages)
        try:
            client = self.get_anthropic_client(provider_config.anthropic)
            completion = client.completions.create(
                prompt=formatted_prompt,
                **self.get_model_params(llm_parameters)
            )
            content = completion.completion
            message_history = messages + [{"role": "assistant", "content": content}]
            return ChatCompletionResponse(
                content=content,
                is_complete=completion.stop_reason == 'stop_sequence',
                message_history=message_history,
            )
        except anthropic.APIError as e:
            raise FreeplayError("Error calling Anthropic") from e

    def continue_chat_stream(
            self,
            messages: List[ChatMessage],
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> Generator[CompletionChunk, None, None]:
        formatted_prompt = self.__to_anthropic_chat_format(messages)
        try:
            client = self.get_anthropic_client(provider_config.anthropic)
            anthropic_response = client.completions.create(
                prompt=formatted_prompt,
                stream=True,
                **self.get_model_params(llm_parameters)
            )

            for chunk in anthropic_response:
                yield CompletionChunk(
                    text=chunk.completion,
                    is_complete=chunk.stop_reason == 'stop_sequence'
                )
        except anthropic.APIError as e:
            raise FreeplayError("Error calling Anthropic") from e

    def call_service(self, formatted_prompt: str, provider_config: ProviderConfig,
                     llm_parameters: LLMParameters) -> CompletionResponse:
        messages = json.loads(formatted_prompt)
        completion = self.continue_chat(messages, provider_config, llm_parameters)
        return CompletionResponse(
            content=completion.content,
            is_complete=completion.is_complete,
        )

    def call_service_stream(
            self,
            formatted_prompt: str,
            provider_config: ProviderConfig,
            llm_parameters: LLMParameters
    ) -> Generator[CompletionChunk, None, None]:
        messages = json.loads(formatted_prompt)
        return self.continue_chat_stream(messages, provider_config, llm_parameters)


def pick_flavor_from_config(completion_flavor: Optional[Flavor], ui_flavor_name: Optional[str]) -> Flavor:
    ui_flavor = Flavor.get_by_name(ui_flavor_name) if ui_flavor_name else None
    flavor = completion_flavor or ui_flavor

    if flavor is None:
        raise FreeplayConfigurationError(
            "Flavor must be configured on either the Freeplay client, completion call, "
            "or in the Freeplay UI. Unable to fulfill request.")

    return flavor


def get_chat_flavor_from_config(completion_flavor: Optional[Flavor], ui_flavor_name: Optional[str]) -> ChatFlavor:
    flavor = pick_flavor_from_config(completion_flavor, ui_flavor_name)
    return require_chat_flavor(flavor)


def require_chat_flavor(flavor: Flavor) -> ChatFlavor:
    if not isinstance(flavor, ChatFlavor):
        raise FreeplayConfigurationError('A Chat flavor is required to start a chat session.')

    return flavor



from .freeplay_thin import Freeplay
from .model import CallInfo, ResponseInfo, RecordPayload, PromptInfo, TestRunInfo

__all__ = [
    'Freeplay',
    'CallInfo',
    'PromptInfo',
    'RecordPayload',
    'ResponseInfo',
    'TestRunInfo',
]



from freeplay.support import CallSupport
from freeplay.thin.model import TestRun, TestCase


class TestRuns:
    def __init__(self, call_support: CallSupport) -> None:
        self.call_support = call_support

    def create(self, project_id: str, testlist: str) -> TestRun:
        test_run = self.call_support.create_test_run(project_id, testlist)
        test_cases = [
            TestCase(test_case_id=test_case.id, variables=test_case.variables)
            for test_case in test_run.test_cases
        ]

        return TestRun(test_run.test_run_id, test_cases)



from dataclasses import dataclass
from typing import Optional, List, cast, Union

from freeplay.completions import OpenAIFunctionCall, ChatMessage
from freeplay.flavors import Flavor
from freeplay.llm_parameters import LLMParameters
from freeplay.model import InputVariables
from freeplay.utils import bind_template_variables


@dataclass
class CallInfo:
    provider: str
    model: str
    start_time: float
    end_time: float
    model_parameters: LLMParameters


@dataclass
class PromptInfo:
    prompt_template_id: str
    prompt_template_version_id: str
    template_name: str
    environment: str
    model_parameters: LLMParameters
    provider: str
    model: str
    flavor_name: str

    def get_call_info(self, start_time: float, end_time: float) -> CallInfo:
        return CallInfo(
            self.provider,
            self.model,
            start_time,
            end_time,
            self.model_parameters
        )


@dataclass
class ResponseInfo:
    is_complete: bool
    function_call_response: Optional[OpenAIFunctionCall] = None
    prompt_tokens: Optional[int] = None
    response_tokens: Optional[int] = None


@dataclass
class TestRunInfo:
    test_run_id: str
    test_case_id: str


@dataclass
class RecordPayload:
    all_messages: List[dict[str, str]]
    inputs: InputVariables
    session_id: str

    prompt_info: PromptInfo
    call_info: CallInfo
    response_info: ResponseInfo
    test_run_info: Optional[TestRunInfo] = None


@dataclass
class Session:
    session_id: str


class FormattedPrompt:
    def __init__(
            self,
            prompt_info: PromptInfo,
            messages: List[dict[str, str]],
            formatted_prompt: Union[str, List[dict[str, str]]]
    ):
        self.prompt_info = prompt_info
        self.messages = messages
        self.llm_prompt = formatted_prompt

    def all_messages(
            self,
            new_message: dict[str, str]
    ) -> List[dict[str, str]]:
        return self.messages + [new_message]


class BoundPrompt:
    def __init__(
            self,
            prompt_info: PromptInfo,
            messages: List[dict[str, str]]
    ):
        self.prompt_info = prompt_info
        self.messages = messages

    def format(
            self,
            flavor_name: Optional[str] = None
    ) -> FormattedPrompt:
        final_flavor = flavor_name or self.prompt_info.flavor_name
        flavor = Flavor.get_by_name(final_flavor)
        llm_format = flavor.to_llm_syntax(cast(List[ChatMessage], self.messages))

        return FormattedPrompt(
            self.prompt_info,
            self.messages,
            cast(Union[str, List[dict[str, str]]], llm_format)
        )


class TemplatePrompt:
    def __init__(
            self,
            prompt_info: PromptInfo,
            messages: List[dict[str, str]]
    ):
        self.prompt_info = prompt_info
        self.messages = messages

    def bind(self, variables: InputVariables) -> BoundPrompt:
        bound_messages = [
            {'role': message['role'], 'content': bind_template_variables(message['content'], variables)}
            for message in self.messages
        ]
        return BoundPrompt(self.prompt_info, bound_messages)


@dataclass
class TestCase:
    def __init__(self, test_case_id: str, variables: InputVariables):
        self.id = test_case_id
        self.variables = variables


@dataclass
class TestRun:
    def __init__(
            self,
            test_run_id: str,
            test_cases: List[TestCase]
    ):
        self.test_run_id = test_run_id
        self.test_cases = test_cases

    def get_test_cases(self) -> List[TestCase]:
        return self.test_cases

    def get_test_run_info(self, test_case_id: str) -> TestRunInfo:
        return TestRunInfo(self.test_run_id, test_case_id)



from dataclasses import dataclass
from typing import List, Union, Any, Dict, Mapping

from pydantic import RootModel

InputValue = Union[str, int, bool, dict[str, Any], list[Any]]
InputVariable = RootModel[Union[Dict[str, "InputVariable"], List["InputVariable"], str, int, bool, float]]
InputVariable.model_rebuild()

InputVariables = Mapping[str, InputValue]

PydanticInputVariables = RootModel[Dict[str, InputVariable]]

TestRunInput = Mapping[str, InputValue]


@dataclass
class TestRun:
    id: str
    inputs: List[TestRunInput]



import json

from freeplay.completions import PromptTemplateWithMetadata
from freeplay.errors import FreeplayClientError
from freeplay.record import RecordCallFields
from freeplay.support import CallSupport
from freeplay.thin.model import RecordPayload


class Recordings:
    def __init__(self, call_support: CallSupport):
        self.call_support = call_support

    def create(self, record_payload: RecordPayload) -> None:
        if len(record_payload.all_messages) < 1:
            raise FreeplayClientError("Messages list must have at least one message. "
                                      "The last message should be the current response.")

        completion = record_payload.all_messages[-1]
        history_as_string = json.dumps(record_payload.all_messages[0:-1])

        template = PromptTemplateWithMetadata(
            prompt_template_id=record_payload.prompt_info.prompt_template_id,
            prompt_template_version_id=record_payload.prompt_info.prompt_template_version_id,
            name=record_payload.prompt_info.template_name,
            content=history_as_string,
            flavor_name=record_payload.prompt_info.flavor_name,
            params=record_payload.prompt_info.model_parameters
        )

        self.call_support.record_processor.record_call(
            RecordCallFields(
                formatted_prompt=history_as_string,
                completion_content=completion['content'],
                completion_is_complete=record_payload.response_info.is_complete,
                start=record_payload.call_info.start_time,
                end=record_payload.call_info.end_time,
                session_id=record_payload.session_id,
                target_template=template,
                variables=record_payload.inputs,
                tag=record_payload.prompt_info.environment,
                test_run_id=record_payload.test_run_info.test_run_id if record_payload.test_run_info else None,
                test_case_id=record_payload.test_run_info.test_case_id if record_payload.test_run_info else None,
                model=record_payload.call_info.model,
                provider=record_payload.prompt_info.provider,
                llm_parameters=record_payload.call_info.model_parameters,
                record_format_type=None  # This is deprecated and unused in the API
            )
        )


import logging
from dataclasses import dataclass
from typing import Dict, Optional, Union
from abc import abstractmethod, ABC
from dataclasses import dataclass
from typing import Optional

from . import api_support
from .completions import PromptTemplateWithMetadata
from .llm_parameters import LLMParameters
from .model import InputVariables

logger = logging.getLogger(__name__)


@dataclass
class RecordCallFields:
    completion_content: str
    completion_is_complete: bool
    end: float
    formatted_prompt: str
    session_id: str
    start: float
    target_template: PromptTemplateWithMetadata
    variables: InputVariables
    tag: str
    test_run_id: Optional[str]
    test_case_id: Optional[str]
    record_format_type: Optional[str]
    model: Optional[str]
    provider: Optional[str]
    llm_parameters: Optional[LLMParameters]
    custom_metadata: Optional[Dict[str, Union[str,int,float]]] = None


class RecordProcessor(ABC):
    @abstractmethod
    def record_call(
            self,
            record_call: RecordCallFields,
    ) -> None:
        pass


class NoOpRecorder(RecordProcessor):
    def record_call(
            self,
            record_call: RecordCallFields,
    ) -> None:
        pass


no_op_recorder = NoOpRecorder()


class DefaultRecordProcessor(RecordProcessor):

    def __init__(
            self,
            freeplay_api_key: str,
            api_base: str
    ) -> None:
        self.api_base = api_base
        self.freeplay_api_key = freeplay_api_key

    def record_call(
            self,
            record_call: RecordCallFields
    ) -> None:
        record_payload = {
            "session_id": record_call.session_id,
            "project_version_id": record_call.target_template.prompt_template_version_id,
            "prompt_template_id": record_call.target_template.prompt_template_id,
            "start_time": record_call.start,
            "end_time": record_call.end,
            "tag": record_call.tag,
            "inputs": record_call.variables,
            "prompt_content": record_call.formatted_prompt,
            "return_content": record_call.completion_content,
            "format_type": record_call.record_format_type,
            "is_complete": record_call.completion_is_complete,
            "model": record_call.model,
            "provider": record_call.provider,
            "llm_parameters": record_call.llm_parameters,
            "custom_metadata": record_call.custom_metadata,
        }

        if record_call.test_run_id is not None:
            record_payload['test_run_id'] = record_call.test_run_id

        if record_call.test_case_id is not None:
            record_payload['test_case_id'] = record_call.test_case_id

        try:
            recorded_response = api_support.post_raw(
                api_key=self.freeplay_api_key,
                url=f'{self.api_base}/v1/record',
                payload=record_payload
            )
            recorded_response.raise_for_status()
        except Exception as e:
            status_code = -1
            if hasattr(e, 'response') and hasattr(e.response, 'status_code'):
                status_code = e.response.status_code
            logger.warning(
                f'There was an error recording to Freeplay. Call will not be logged. '
                f'Status: {status_code}. {e.__class__}'
            )



import json
from typing import Optional

from freeplay.completions import PromptTemplates
from freeplay.errors import FreeplayConfigurationError
from freeplay.flavors import Flavor
from freeplay.model import InputVariables
from freeplay.support import CallSupport
from freeplay.thin.model import TemplatePrompt, PromptInfo, FormattedPrompt


class Prompts:
    def __init__(self, call_support: CallSupport) -> None:
        self.call_support = call_support

    def get_all(self, project_id: str, environment: str) -> PromptTemplates:
        return self.call_support.get_prompts(project_id=project_id, tag=environment)

    def get(self, project_id: str, template_name: str, environment: str) -> TemplatePrompt:
        prompt_template = self.call_support.get_prompt(
            project_id=project_id,
            template_name=template_name,
            environment=environment
        )

        messages = json.loads(prompt_template.content)

        params = prompt_template.get_params()
        model = params.pop('model')

        if not prompt_template.flavor_name:
            raise FreeplayConfigurationError(
                "Flavor must be configured in the Freeplay UI. Unable to fulfill request.")

        flavor = Flavor.get_by_name(prompt_template.flavor_name)

        prompt_info = PromptInfo(
            prompt_template_id=prompt_template.prompt_template_id,
            prompt_template_version_id=prompt_template.prompt_template_version_id,
            template_name=prompt_template.name,
            environment=environment,
            model_parameters=params,
            provider=flavor.provider,
            model=model,
            flavor_name=prompt_template.flavor_name
        )

        return TemplatePrompt(prompt_info, messages)

    def get_formatted(
            self,
            project_id: str,
            template_name: str,
            environment: str,
            variables: InputVariables,
            flavor_name: Optional[str] = None
    ) -> FormattedPrompt:
        bound_prompt = self.get(
            project_id=project_id,
            template_name=template_name,
            environment=environment
        ).bind(variables=variables)

        return bound_prompt.format(flavor_name)



import json
import time
from copy import copy
from typing import Dict, Any, Optional, Union, List, Generator
from uuid import uuid4

from freeplay import api_support
from freeplay.api_support import try_decode
from freeplay.completions import PromptTemplates, PromptTemplateWithMetadata, ChatMessage, ChatCompletionResponse, \
    CompletionChunk, CompletionResponse
from freeplay.errors import FreeplayConfigurationError, freeplay_response_error, FreeplayServerError
from freeplay.flavors import ChatFlavor, Flavor, pick_flavor_from_config
from freeplay.llm_parameters import LLMParameters
from freeplay.model import InputVariables
from freeplay.provider_config import ProviderConfig
from freeplay.record import RecordProcessor, RecordCallFields

JsonDom = Dict[str, Any]


class TestCaseTestRunResponse:
    def __init__(self, test_case: JsonDom):
        self.variables: InputVariables = test_case['variables']
        self.id: str = test_case['id']


class TestRunResponse:
    def __init__(
            self,
            test_run_id: str,
            test_cases: list[JsonDom]
    ):
        self.test_cases = [
            TestCaseTestRunResponse(test_case)
            for test_case in test_cases
        ]
        self.test_run_id = test_run_id


class CallSupport:
    def __init__(
            self,
            freeplay_api_key: str,
            api_base: str,
            record_processor: RecordProcessor,
            **kwargs: Any
    ) -> None:
        self.api_base = api_base
        self.freeplay_api_key = freeplay_api_key
        self.client_params = LLMParameters(kwargs)
        self.record_processor = record_processor

    @staticmethod
    def find_template_by_name(prompts: PromptTemplates, template_name: str) -> PromptTemplateWithMetadata:
        templates = [t for t in prompts.templates if t.name == template_name]
        if len(templates) == 0:
            raise FreeplayConfigurationError(f'Could not find template with name "{template_name}"')
        return templates[0]

    def create_session_id(self) -> str:
        return str(uuid4())

    def check_all_values_string_or_number(self, metadata: Optional[Dict[str, Union[str, int, float]]]) -> None:
        if metadata:
            for key, value in metadata.items():
                if not isinstance(value, (str, int, float)):
                    raise FreeplayConfigurationError(f"Invalid value for key {key}: Value must be a string or number.")

    def get_prompt(self, project_id: str, template_name: str, environment: str) -> PromptTemplateWithMetadata:
        prompt_templates = self.get_prompts(project_id, environment)
        return self.find_template_by_name(prompt_templates, template_name)

    def get_prompts(self, project_id: str, tag: str) -> PromptTemplates:
        response = api_support.get_raw(
            api_key=self.freeplay_api_key,
            url=f'{self.api_base}/projects/{project_id}/templates/all/{tag}'
        )

        if response.status_code != 200:
            raise freeplay_response_error("Error getting prompt templates", response)

        maybe_prompts = try_decode(PromptTemplates, response.content)
        if maybe_prompts is None:
            raise FreeplayServerError(f'Failed to parse prompt templates from server')

        return maybe_prompts

    def create_test_run(self, project_id: str, testlist: str) -> TestRunResponse:
        response = api_support.post_raw(
            api_key=self.freeplay_api_key,
            url=f'{self.api_base}/projects/{project_id}/test-runs-cases',
            payload={'testlist_name': testlist},
        )

        if response.status_code != 201:
            raise freeplay_response_error('Error while creating a test run.', response)

        json_dom = response.json()

        return TestRunResponse(json_dom['test_run_id'], json_dom['test_cases'])

    # noinspection PyUnboundLocalVariable
    def prepare_and_make_chat_call(
            self,
            session_id: str,
            flavor: ChatFlavor,
            provider_config: ProviderConfig,
            tag: str,
            target_template: PromptTemplateWithMetadata,
            variables: InputVariables,
            message_history: List[ChatMessage],
            new_messages: Optional[List[ChatMessage]],
            test_run_id: Optional[str] = None,
            completion_parameters: Optional[LLMParameters] = None,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None
    ) -> ChatCompletionResponse:
        # make call
        start = time.time()
        params = target_template.get_params() \
            .merge_and_override(self.client_params) \
            .merge_and_override(completion_parameters)
        prompt_messages = copy(message_history)
        if new_messages is not None:
            prompt_messages.extend(new_messages)
        completion_response = flavor.continue_chat(messages=prompt_messages,
                                                   provider_config=provider_config,
                                                   llm_parameters=params)
        end = time.time()

        model = flavor.get_model_params(params).get('model')
        formatted_prompt = json.dumps(prompt_messages)
        # record data
        record_call_fields = RecordCallFields(
            completion_content=completion_response.content,
            completion_is_complete=completion_response.is_complete,
            end=end,
            formatted_prompt=formatted_prompt,
            session_id=session_id,
            start=start,
            target_template=target_template,
            variables=variables,
            record_format_type=flavor.record_format_type,
            tag=tag,
            test_run_id=test_run_id,
            test_case_id=None,
            model=model,
            provider=flavor.provider,
            llm_parameters=params,
            custom_metadata=metadata,
        )
        self.record_processor.record_call(record_call_fields)

        return completion_response

    # noinspection PyUnboundLocalVariable
    def prepare_and_make_chat_call_stream(
            self,
            session_id: str,
            flavor: ChatFlavor,
            provider_config: ProviderConfig,
            tag: str,
            target_template: PromptTemplateWithMetadata,
            variables: InputVariables,
            message_history: List[ChatMessage],
            test_run_id: Optional[str] = None,
            completion_parameters: Optional[LLMParameters] = None,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None
    ) -> Generator[CompletionChunk, None, None]:
        # make call
        start = time.time()
        prompt_messages = copy(message_history)
        params = target_template.get_params() \
            .merge_and_override(self.client_params) \
            .merge_and_override(completion_parameters)
        completion_response = flavor.continue_chat_stream(prompt_messages, provider_config, llm_parameters=params)

        str_content = ''
        last_is_complete = False
        for chunk in completion_response:
            str_content += chunk.text or ''
            last_is_complete = chunk.is_complete
            yield chunk
        # End time must be logged /after/ streaming the response above, or else OpenAI latency will not be captured.
        end = time.time()

        model = flavor.get_model_params(params).get('model')
        formatted_prompt = json.dumps(prompt_messages)
        record_call_fields = RecordCallFields(
            completion_content=str_content,
            completion_is_complete=last_is_complete,
            end=end,
            formatted_prompt=formatted_prompt,
            session_id=session_id,
            start=start,
            target_template=target_template,
            variables=variables,
            record_format_type=flavor.record_format_type,
            tag=tag,
            test_run_id=test_run_id,
            test_case_id=None,
            model=model,
            provider=flavor.provider,
            llm_parameters=params,
            custom_metadata=metadata,
        )
        self.record_processor.record_call(record_call_fields)

    # noinspection PyUnboundLocalVariable
    def prepare_and_make_call(
            self,
            session_id: str,
            prompts: PromptTemplates,
            template_name: str,
            variables: InputVariables,
            flavor: Optional[Flavor],
            provider_config: ProviderConfig,
            tag: str,
            test_run_id: Optional[str] = None,
            completion_parameters: Optional[LLMParameters] = None,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None
    ) -> CompletionResponse:
        target_template = self.find_template_by_name(prompts, template_name)
        params = target_template.get_params() \
            .merge_and_override(self.client_params) \
            .merge_and_override(completion_parameters)

        final_flavor = pick_flavor_from_config(flavor, target_template.flavor_name)
        formatted_prompt = final_flavor.format(target_template, variables)

        # make call
        start = time.time()
        completion_response = final_flavor.call_service(formatted_prompt=formatted_prompt,
                                                        provider_config=provider_config,
                                                        llm_parameters=params)
        end = time.time()

        model = final_flavor.get_model_params(params).get('model')

        # record data
        record_call_fields = RecordCallFields(
            completion_content=completion_response.content,
            completion_is_complete=completion_response.is_complete,
            end=end,
            formatted_prompt=formatted_prompt,
            session_id=session_id,
            start=start,
            target_template=target_template,
            variables=variables,
            record_format_type=final_flavor.record_format_type,
            tag=tag,
            test_run_id=test_run_id,
            test_case_id=None,
            model=model,
            provider=final_flavor.provider,
            llm_parameters=params,
            custom_metadata=metadata,
        )
        self.record_processor.record_call(record_call_fields)

        return completion_response

    def prepare_and_make_call_stream(
            self,
            session_id: str,
            prompts: PromptTemplates,
            template_name: str,
            variables: InputVariables,
            flavor: Optional[Flavor],
            provider_config: ProviderConfig,
            tag: str,
            test_run_id: Optional[str] = None,
            completion_parameters: Optional[LLMParameters] = None,
            metadata: Optional[Dict[str, Union[str, int, float]]] = None
    ) -> Generator[CompletionChunk, None, None]:
        target_template = self.find_template_by_name(prompts, template_name)
        params = target_template.get_params() \
            .merge_and_override(self.client_params) \
            .merge_and_override(completion_parameters)

        final_flavor = pick_flavor_from_config(flavor, target_template.flavor_name)
        formatted_prompt = final_flavor.format(target_template, variables)

        # make call
        start = int(time.time())
        completion_response = final_flavor.call_service_stream(
            formatted_prompt=formatted_prompt, provider_config=provider_config, llm_parameters=params)
        text_chunks = []
        last_is_complete = False
        for chunk in completion_response:
            text_chunks.append(chunk.text)
            last_is_complete = chunk.is_complete
            yield chunk
        # End time must be logged /after/ streaming the response above, or else OpenAI latency will not be captured.
        end = int(time.time())

        model = final_flavor.get_model_params(params).get('model')

        record_call_fields = RecordCallFields(
            completion_content=''.join(text_chunks),
            completion_is_complete=last_is_complete,
            end=end,
            formatted_prompt=formatted_prompt,
            session_id=session_id,
            start=start,
            target_template=target_template,
            variables=variables,
            record_format_type=final_flavor.record_format_type,
            tag=tag,
            test_run_id=test_run_id,
            test_case_id=None,
            model=model,
            provider=final_flavor.provider,
            llm_parameters=params,
            custom_metadata=metadata,
        )
        self.record_processor.record_call(record_call_fields)



from typing import Dict, Union, Optional
import importlib.metadata
import platform

import pystache  # type: ignore
from pydantic import ValidationError

from .errors import FreeplayError, FreeplayConfigurationError
from .model import PydanticInputVariables, InputVariables


def bind_template_variables(template: str, variables: InputVariables) -> str:
    # Validate that the variables are of the correct type, and do not include functions or None values.
    try:
        PydanticInputVariables.model_validate(variables)
    except ValidationError as err:
        raise FreeplayError(
            'Variables must be a string, number, bool, or a possibly nested'
            ' list or dict of strings, numbers and booleans.'
        )

    # When rendering mustache, do not escape HTML special characters.
    rendered: str = pystache.Renderer(escape=lambda s: s).render(template, variables)
    return rendered


def check_all_values_string_or_number(metadata: Optional[Dict[str, Union[str, int, float]]]) -> None:
    if metadata:
        for key, value in metadata.items():
            if not isinstance(value, (str, int, float)):
                raise FreeplayConfigurationError(f"Invalid value for key {key}: Value must be a string or number.")


def build_request_header(api_key: str) -> Dict[str, str]:
    return {
        'Authorization': f'Bearer {api_key}',
        'User-Agent': get_user_agent()
    }


def get_user_agent() -> str:
    sdk_name = 'Freeplay'
    sdk_version = importlib.metadata.version('Freeplay')
    language = 'Python'
    language_version = platform.python_version()
    os_name = platform.system()
    os_version = platform.release()

    # Output format
    # Freeplay/0.2.30 (Python/3.11.4; Darwin/23.2.0)
    return f"{sdk_name}/{sdk_version} ({language}/{language_version}; {os_name}/{os_version})"
